# Implementation of the single shooting scheme of optimal control problems
#close all
import casadi as C
import numpy as np
import pylab


# Create NLP
N   = 30       # Control discretization
T   = 3.0      # End time
h   = T/N      # Discretization step
nx  = 1            
nu  = 1

Q = 1
R = 1

# Declare variables (use scalar graph)
u  = C.SX.sym('u',nu)    # control
x  = C.SX.sym('x',nx)    # states

# System dynamics
xdotl= -x[0]  + u[0]
xdotr= -(1 + x[0])*(x[0])  + u[0]
f = C.Function('f', [x,u],[xdotl])
fr = C.Function('fr', [x,u],[xdotr])

# RK4
U   = C.MX.sym('U')
X0  = C.MX.sym('X0',nx)
X   = X0

[k1] = f([X,U])
[k2] = f([X + h/2 * k1, U])
[k3] = f([X + h/2 * k2, U])
[k4] = f([X + h   * k3, U])

X = X + h/6*(k1   + 2*k2   + 2*k3   + k4)
RK4 = C.Function('RK4',[X0,U],[X])


[k1] = fr([X,U])
[k2] = fr([X + h/2 * k1, U])
[k3] = fr([X + h/2 * k2, U])
[k4] = fr([X + h   * k3, U])

X = X + h/6*(k1   + 2*k2   + 2*k3   + k4)
RK4r = C.Function('RK4r',[X0,U],[X])

# Formulate NLP 
u = C.SX.sym('u',N)

# Objective function
J=0

# Get an expression for the cost and state at end

x_0 = 1
X = x_0
g = []; lbg = []; ubg = [];
min_x = 0.5
for i in range(N):
    [X] = RK4([X,u[i]])
    if (i >= 14 and i <= 16):
        g = C.vertcat([g, X])
        ubg = C.vertcat([ubg, np.Inf])
        lbg = C.vertcat([lbg, min_x])
    J = J + Q*X[0]**2 + R*u[i]**2

# Terminal constraints: x_0(T)=x_1(T)=0
g = C.vertcat([g, X])
ubg = C.vertcat([ubg, 0])
lbg = C.vertcat([lbg, 0])

# Allocate an NLP solver
nlp = {'x': u, 'f': J, 'g': g}

# Create IPOPT solver object
options = {"ipopt": {"hessian_approximation":"limited-memory"}}
solver = C.nlpsol('solver', 'ipopt', nlp, options)

# Solve the NLP
arg = {}
arg["x0"]  = 0.    # solution guess
arg["lbx"] = -2    # lower bound on x
arg["ubx"] =  2    # upper bound on x
arg["lbg"] =  lbg    # lower bound on g
arg["ubg"] =  ubg    # upper bound on g

# Solve the problem
res = solver(arg)

f_opt       = res["f"].full()
u_opt       = res["x"].full()

# Compute state trajectory integrating the dynamics
x_opt       = np.zeros((nx,N+1))
x_optr      = np.zeros((nx,N+1))
x_lqr       = np.zeros((nx,N+1))
x_opt[:,0]  = x_0
x_optr[:,0] = x_0
x_lqr[:,0]  = x_0

for i in range(1,N):
    out         = RK4([x_opt[:,i-1],u_opt[i-1]])
    x_opt[:,i]    = out[0].full()
    out         = RK4r([x_optr[:,i-1],u_opt[i-1]])
    x_optr[:,i]   = out[0].full()
    out         = RK4r([x_lqr[:,i-1],-0.0501*x_lqr[:,i-1]])
    x_lqr[:,i] = out[0].full()

# Plot results

pylab.figure()
pylab.subplot(2,1,1)
pylab.plot(x_opt.T)
pylab.plot(x_optr.T)
pylab.plot(x_lqr.T)
pylab.xlabel('time - t')
pylab.ylabel('state - x')
pylab.grid(True)
pylab.legend(('exact','linearized','lqr'))
pylab.subplot(2,1,2)
pylab.plot(u_opt,'r')
pylab.xlabel('time - t')
pylab.ylabel('input - u')
pylab.grid(True)

pylab.show()

